/*
 * xfrm4_mode_beet.c - BEET mode encapsulation for IPv4.
 *
 * Copyright (c) 2006 Diego Beltrami <diego.beltrami@gmail.com>
 *                    Miika Komu     <miika@iki.fi>
 *                    Herbert Xu     <herbert@gondor.apana.org.au>
 *                    Abhinav Pathak <abhinav.pathak@hiit.fi>
 *                    Jeff Ahrenholz <ahrenholz@gmail.com>
 *                    Joakim Koskela <jookos@gmail.com>
 */

#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/stringify.h>
#include <net/dst.h>
#include <net/ip.h>
#include <net/xfrm.h>

/* Add encapsulation header.
 *
 * The top IP header will be constructed per draft-nikander-esp-beet-mode-06.txt.
 * The following fields in it shall be filled in by x->type->output:
 *      tot_len
 *      check
 *
 * On exit, skb->h will be set to the start of the payload to be processed
 * by x->type->output and skb->nh will be set to the top IP header.
 */
static int xfrm4_beet_output(struct xfrm_state *x, struct sk_buff *skb)
{
        struct dst_entry *dst = skb->dst;
        int hdrlen;
        struct iphdr *iphv4, *top_iphv4;
        struct ipv6hdr *iphv6, *top_iphv6;

        if (skb->nh.iph->version == 4) {
                
                int optlen;

                /* 4-4 */

                iphv4 = skb->nh.iph;
                skb->h.ipiph = iphv4;
                
                hdrlen = x->props.header_len;

                optlen = iphv4->ihl * 4 - sizeof(*iphv4);

                if (!optlen) {
                        hdrlen -= IPV4_BEET_PHMAXLEN;
                } else {
                        skb->h.raw -= (IPV4_BEET_PHMAXLEN - (optlen & 4));
                        hdrlen -= optlen & 4;
                }
                
                skb->nh.raw = skb_push(skb, hdrlen);
                                
                top_iphv4 = skb->nh.iph;
                hdrlen = iphv4->ihl * 4 - optlen;
                skb->h.raw += hdrlen;
                memmove(top_iphv4, iphv4, hdrlen);
                
                if (unlikely(optlen)) {
                        struct ip_beet_phdr *ph;
                        
                        BUG_ON(optlen < 0);
                        
                        ph = (struct ip_beet_phdr *)skb->h.raw;
                        ph->padlen = 4 - (optlen & 4);
                        ph->hdrlen = (optlen + ph->padlen + sizeof(*ph)) / 8;
                        ph->nexthdr = iphv4->protocol;
                        top_iphv4->protocol = IPPROTO_BEETPH;
                        top_iphv4->ihl = sizeof(struct iphdr) / 4;
                }
                
                top_iphv4->saddr = x->props.saddr.a4;
                top_iphv4->daddr = x->id.daddr.a4;
                
                skb->protocol = htons(ETH_P_IP);

	} else if (skb->nh.iph->version == 6) {
                
                u8 protocol;
                int delta = sizeof(struct ipv6hdr) - sizeof(struct iphdr);

                /* Inner = 6, Outer = 4 : changing the external IP hdr
                 * to the outer addresses
                 */

                hdrlen = x->props.header_len - IPV4_BEET_PHMAXLEN;
                skb_push(skb, hdrlen);
                iphv6 = skb->nh.ipv6h;
                
                skb->nh.raw = skb->data;
                top_iphv6 = skb->nh.ipv6h;

                protocol = iphv6->nexthdr;
                skb->nh.raw = skb_pull(skb, delta);
                top_iphv4 = skb->nh.iph;
                skb->h.raw = skb->data + hdrlen;
                top_iphv4->ihl = (sizeof(struct iphdr) >> 2);
                top_iphv4->version = 4;
                top_iphv4->id = 0;
                top_iphv4->frag_off = htons(IP_DF);
                top_iphv4->ttl = dst_metric(dst->child, RTAX_HOPLIMIT);
                top_iphv4->saddr = x->props.saddr.a4;
                top_iphv4->daddr = x->id.daddr.a4;
                skb->h.raw += top_iphv4->ihl*4;
                top_iphv4->protocol = protocol;
         
                skb->protocol = htons(ETH_P_IP);
        } else 
                BUG_ON(1);
        
	return 0; 
}

static int xfrm4_beet_input(struct xfrm_state *x, struct sk_buff *skb)
{
	struct iphdr *iph = skb->nh.iph;
	int hops = skb->nh.iph->ttl;
	int phlen = 0;
	int optlen = 0;
	int size = (x->sel.family == AF_INET) ? sizeof(struct iphdr) : sizeof(struct ipv6hdr);
	int delta = sizeof(struct ipv6hdr) - sizeof(struct iphdr);
	__u8 ph_nexthdr = 0, protocol = 0;
	int err = -EINVAL;

	protocol = iph->protocol;

	if (x->sel.family == AF_INET6) {
		/* Here, the inner family is 6, therefore I have to
		 * substitute the IPhdr by enlarging it */
		if (skb_tailroom(skb) <  delta){
			if (pskb_expand_head(skb, 0, delta, GFP_ATOMIC))
				goto out;
		}
		skb->nh.raw -= delta;
	} else if (x->sel.family == AF_INET) {
		if (unlikely(iph->protocol == IPPROTO_BEETPH)) {
			struct ip_beet_phdr *ph = (struct ip_beet_phdr*)(iph + 1);

			if (!pskb_may_pull(skb, sizeof(*ph)))
				goto out;

			phlen = ph->hdrlen * 8;
			optlen = phlen - ph->padlen - sizeof(*ph);
			if (optlen < 0 || optlen & 3 || optlen > 250)
				goto out;

			if (!pskb_may_pull(skb, phlen))
				goto out;

			ph_nexthdr = ph->nexthdr;
		}
	} else
		BUG_ON(1);

	size += (optlen - phlen);
	skb_push(skb, size);
	memmove(skb->data, skb->nh.raw, sizeof(*iph));
	skb->nh.raw = skb->data;

	if (x->sel.family == AF_INET) {
		iph = skb->nh.iph;
		iph->ihl = (sizeof(*iph) + optlen) / 4;
		iph->tot_len = htons(skb->len);
		iph->daddr = x->sel.daddr.a4;
		iph->saddr = x->sel.saddr.a4;
		if (ph_nexthdr)
			iph->protocol = ph_nexthdr;
		else
			iph->protocol = protocol;
		iph->check = 0;
		iph->check = ip_fast_csum(skb->nh.raw, iph->ihl);
	} else if (x->sel.family == AF_INET6) {
		struct ipv6hdr *ip6h = skb->nh.ipv6h;
		memset(ip6h->flow_lbl, 0, sizeof(ip6h->flow_lbl));
		ip6h->version = 6;
		ip6h->priority = 0;
		ip6h->nexthdr = protocol;
		ip6h->hop_limit = hops;
		ip6h->payload_len = htons(skb->len - size);
		ipv6_addr_copy(&ip6h->daddr, (struct in6_addr *)&x->sel.daddr.a6);
		ipv6_addr_copy(&ip6h->saddr, (struct in6_addr *)&x->sel.saddr.a6);
		skb->protocol = htons(ETH_P_IPV6);
	}
	err = 0;
out:
	return err;
}

static struct xfrm_mode xfrm4_beet_mode = {
	.input = xfrm4_beet_input,
	.output = xfrm4_beet_output,
	.owner = THIS_MODULE,
	.encap = XFRM_MODE_BEET,
};

static int __init xfrm4_beet_init(void)
{
	return xfrm_register_mode(&xfrm4_beet_mode, AF_INET);
}

static void __exit xfrm4_beet_exit(void)
{
	int err;

	err = xfrm_unregister_mode(&xfrm4_beet_mode, AF_INET);
	BUG_ON(err);
}

module_init(xfrm4_beet_init);
module_exit(xfrm4_beet_exit);
MODULE_LICENSE("GPL");
MODULE_ALIAS_XFRM_MODE(AF_INET, XFRM_MODE_BEET);
