/*
 * xfrm6_mode_beet.c - BEET mode encapsulation for IPv6.
 *
 * Copyright (c) 2006 Diego Beltrami <diego.beltrami@gmail.com>
 *                    Miika Komu     <miika@iki.fi>
 *                    Herbert Xu     <herbert@gondor.apana.org.au>
 *                    Abhinav Pathak <abhinav.pathak@hiit.fi>
 *                    Jeff Ahrenholz <ahrenholz@gmail.com>
 *                    Joakim Koskela <jookos@gmail.com>
 */

#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/stringify.h>
#include <net/dsfield.h>
#include <net/dst.h>
#include <net/inet_ecn.h>
#include <net/ipv6.h>
#include <net/xfrm.h>

/* Add encapsulation header.
 *
 * The top IP header will be constructed per draft-nikander-esp-beet-mode-06.txt.
 * The following fields in it shall be filled in by x->type->output:
 *	payload_len
 *
 * On exit, skb->h will be set to the start of the encapsulation header to be
 * filled in by x->type->output and skb->nh will be set to the nextheader field
 * of the extension header directly preceding the encapsulation header, or in
 * its absence, that of the top IP header.  The value of skb->data will always
 * point to the top IP header.
 */
static int xfrm6_beet_output(struct xfrm_state *x, struct sk_buff *skb)
{

        struct dst_entry *dst = skb->dst;
        int hdrlen;
        struct iphdr *iphv4, *top_iphv4;
        struct ipv6hdr *iphv6, *top_iphv6;
	
	if (skb->nh.iph->version == 6) {
                
		u8 *prevhdr;
		int hdr_len;

                /* 6-6 */

                hdrlen = x->props.header_len - IPV4_BEET_PHMAXLEN;
                skb_push(skb, hdrlen);
                iphv6 = skb->nh.ipv6h;

		hdr_len = ip6_find_1stfragopt(skb, &prevhdr);
		skb->nh.raw = prevhdr - hdrlen;
		skb->h.raw = skb->data + hdr_len;
		memmove(skb->data, iphv6, hdr_len);
                
                skb->nh.raw = skb->data;
                top_iphv6 = skb->nh.ipv6h;
		skb->nh.raw = &top_iphv6->nexthdr;
		skb->h.ipv6h = top_iphv6 + 1;
		ipv6_addr_copy(&top_iphv6->saddr, (struct in6_addr *)&x->props.saddr);
		ipv6_addr_copy(&top_iphv6->daddr, (struct in6_addr *)&x->id.daddr);

                skb->protocol = htons(ETH_P_IPV6);
                
	} else if (skb->nh.iph->version == 4) {

                int flags;
                int optlen;
                int dsfield;
                u8 protocol;
                int delta = sizeof(struct ipv6hdr) - sizeof(struct iphdr);

                /* Inner = 4, Outer = 6*/

                iphv4 = skb->nh.iph;
                skb->h.ipiph = iphv4;

                hdrlen = x->props.header_len;

                optlen = iphv4->ihl * 4 - sizeof(*iphv4);
                
                if (!optlen) {
                        hdrlen -= IPV4_BEET_PHMAXLEN;
                } else {
                        skb->h.raw -= (IPV4_BEET_PHMAXLEN - (optlen & 4));
                        hdrlen -= optlen & 4;
                }

                skb->nh.raw = skb_push(skb, hdrlen);
                
                top_iphv4 = skb->nh.iph;
                hdrlen = iphv4->ihl * 4 - optlen;
                skb->h.raw += hdrlen;                   
                if (unlikely(optlen)) {
                        struct ip_beet_phdr *ph;
                
                        BUG_ON(optlen < 0);
                
                        ph = (struct ip_beet_phdr *)skb->h.raw;
                        ph->padlen = 4 - (optlen & 4);
                        ph->hdrlen = (optlen + ph->padlen + sizeof(*ph)) / 8;
                        ph->nexthdr = iphv4->protocol;
                        top_iphv4->protocol = IPPROTO_BEETPH;
                        top_iphv4->ihl = sizeof(struct iphdr) / 4;
                }
                
                if (unlikely(optlen))
                        protocol = top_iphv4->protocol;
                else
                        protocol = iphv4->protocol;
                
                if (skb_headroom(skb) <=  2*delta){
                        if (pskb_expand_head(skb, delta,0, GFP_ATOMIC))
                                return -ENOMEM;
                }
                
                skb->nh.raw = skb_push(skb, delta);

                top_iphv6 = skb->nh.ipv6h;
                skb->h.ipv6h = top_iphv6 + 1;
                /* DS disclosed */
                top_iphv6->version = 6;
                top_iphv6->priority = 0;
                top_iphv6->flow_lbl[0] = 0;
                top_iphv6->flow_lbl[1] = 0;
                top_iphv6->flow_lbl[2] = 0;
                dsfield = ipv6_get_dsfield(top_iphv6);
                dsfield = INET_ECN_encapsulate(dsfield, dsfield);
                flags = x->props.flags;
                if (flags & XFRM_STATE_NOECN)
                        dsfield &= ~INET_ECN_MASK;
                ipv6_change_dsfield(top_iphv6, 0, dsfield);

		top_iphv6->nexthdr = protocol;
                top_iphv6->hop_limit = dst_metric(dst->child, RTAX_HOPLIMIT);
                top_iphv6->payload_len = htons(skb->len - sizeof(struct ipv6hdr));
                ipv6_addr_copy(&top_iphv6->saddr,(struct in6_addr *)&x->props.saddr);
                ipv6_addr_copy(&top_iphv6->daddr, (struct in6_addr *)&x->id.daddr);
                skb->nh.raw = &top_iphv6->nexthdr;

                skb->protocol = htons(ETH_P_IPV6);
	} else
		BUG_ON(1);

	return 0;
}

static int xfrm6_beet_input(struct xfrm_state *x, struct sk_buff *skb)
{
	struct ip_beet_phdr *ph = (struct ip_beet_phdr*)(skb->h.raw);
	int size = (x->sel.family == AF_INET) ? sizeof(struct iphdr) : sizeof(struct ipv6hdr);
	int delta = sizeof(struct ipv6hdr) - sizeof(struct iphdr);
	__u8 proto = skb->nh.ipv6h->nexthdr, hops = skb->nh.ipv6h->hop_limit;
	__u8 ph_nexthdr = 0;
	int phlen = 0;
	int optlen = 0;

	int err = -EINVAL;

	if (x->sel.family == AF_INET) {
		/* Inner = IPv4, therefore the IPhdr must be shrunk */
		/* Inner = 4, Outer = 6 */
		if (unlikely(proto == IPPROTO_BEETPH)) {
			if (!pskb_may_pull(skb, sizeof(*ph)))
				goto out;
			phlen = ph->hdrlen * 8;
			optlen = phlen - ph->padlen - sizeof(*ph);

			if (optlen < 0 || optlen & 3 || optlen > 250)
				goto out;
			if (!pskb_may_pull(skb, phlen))
				goto out;

			proto = ph_nexthdr = ph->nexthdr;
		}
		skb->nh.raw += delta;
	}
        
        if (skb_cloned(skb) &&
            pskb_expand_head(skb, 0, 0, GFP_ATOMIC))
                goto out;

	skb_push(skb, size);
	memmove(skb->data, skb->nh.raw, size);
	skb->nh.raw = skb->data;

	skb->mac.raw = memmove(skb->data - skb->mac_len,
			       skb->mac.raw, skb->mac_len);
	if (unlikely(phlen)) {
		skb_pull(skb, phlen - optlen);
		skb->nh.raw = skb->data;
	}
	if (x->sel.family == AF_INET6) {
		struct ipv6hdr *ip6h = skb->nh.ipv6h;
		ip6h->payload_len = htons(skb->len - size);
		ipv6_addr_copy(&ip6h->daddr, (struct in6_addr *) &x->sel.daddr.a6);
		ipv6_addr_copy(&ip6h->saddr, (struct in6_addr *) &x->sel.saddr.a6);
	} else if (x->sel.family == AF_INET) {
		struct iphdr *iph = skb->nh.iph;
		iph->ihl = (sizeof(*iph) + optlen) / 4;
		iph->version = 4;
		iph->tos = 0;
		iph->id = 0;
		iph->frag_off = 0;
		iph->ttl = hops;
		iph->protocol = proto;
		iph->daddr = x->sel.daddr.a4;
		iph->saddr = x->sel.saddr.a4;
		iph->tot_len = htons(skb->len);
		ip_send_check(iph);
		skb->protocol = htons(ETH_P_IP);
		if (unlikely(!optlen)) {
			skb->h.raw = skb->nh.raw;
		}
		dst_release(skb->dst);
		skb->dst = NULL;
	} else
		BUG_ON(1);
	err = 0;
out:
	return err;
}

static struct xfrm_mode xfrm6_beet_mode = {
	.input = xfrm6_beet_input,
	.output = xfrm6_beet_output,
	.owner = THIS_MODULE,
	.encap = XFRM_MODE_BEET,
};

static int __init xfrm6_beet_init(void)
{
	return xfrm_register_mode(&xfrm6_beet_mode, AF_INET6);
}

static void __exit xfrm6_beet_exit(void)
{
	int err;

	err = xfrm_unregister_mode(&xfrm6_beet_mode, AF_INET6);
	BUG_ON(err);
}

module_init(xfrm6_beet_init);
module_exit(xfrm6_beet_exit);
MODULE_LICENSE("GPL");
MODULE_ALIAS_XFRM_MODE(AF_INET6, XFRM_MODE_BEET);
